#include <string>
#include <cstring>
#include <sstream>
#include <algorithm>

#include "base/CommandLineParser.h"
#include "analysis/DNAVector.h"
#include "aligns/KmerAlignCore.h"

#include <math.h>
#include "analysis/KmerTable.h"
#include "analysis/NonRedKmerTable.h"
#include "util/mutil.h"
#include <omp.h>


static float MIN_KMER_ENTROPY = 1.75;
static float MIN_WELD_ENTROPY = 1.5;  // min entropy for each half of a welding mer (kk)
static bool DEBUG = false;

// print nucleotide sequence 80 chars per line
void PrintSeq(const DNAVector & d) {
    int i;
    for (i=0; i<d.isize(); i++) {
        cout << d[i];
        if ((i+1) % 80 == 0)
            cout << endl;
    }
    cout << endl;
}


// wrapper around an integer vector
class Pool
{
public:
    Pool() {}
    
    void add(int i) {m_index.push_back(i);}
    
    int size() const {return m_index.isize();}
    int get(int i) const {return m_index[i];}
    
    void sortvec() {
        sort(m_index.begin(), m_index.end());
    }
    

private:
    svec<int> m_index;
};

// used to determine complexity of a kmer
float compute_entropy(const DNAVector & kmer) {
    
    map<char,int> char_map;
    
    for (unsigned int i = 0; i < (unsigned int)kmer.isize(); i++) {
        
        char c = kmer[i];
        char_map[c]++;
    }
    
    float entropy = 0;
    
    char nucs[] = { 'G', 'A', 'T', 'C' };
    
    for (unsigned int i = 0; i < 4; i++) {
        
        char nuc = nucs[i];
        
        int count = char_map[nuc];
        
        float prob = (float)count / kmer.isize();
        
        if (prob > 0) {
            float val = prob * log(1/prob)/log((float)2);
            entropy += val;
        }
    }
    
    return(entropy);
}

bool IsSimple(const DNAVector & d) 
{
   
    // compute entropy
    float e = compute_entropy(d);

    //cout << d.AsString() << " has entropy: " << e << endl;
    if (e < MIN_KMER_ENTROPY) {
        
        if (DEBUG) 
            cerr << "-Sequence declared low complexity: " << d.AsString() << " with entropy " << e << endl;
        
        return(true);
    }
    else {
        return false;
    }
    

    /*    old check
    
    int i;
    int k = 0;
    for (i=2; i<d.isize(); i++) {
        if (d[i] == d[i-2])
            k++;
    }
    double ratio = (double)k/(double)d.isize();
    if (ratio > 0.6) {

        if (DEBUG) 
            cerr << "-Sequence declared low complexity: " << d.AsString() << " with entropy " << e << endl;
    
        return true;
    }

    */
   
    
}


bool SimpleHalves(const DNAVector & d) {
    int len = d.isize();
    int mid_pos = int(len/2);

    DNAVector left;
    left.resize(mid_pos);
    for (int i = 0; i < mid_pos; i++) {
        left[i] = d[i];
    }
    
    DNAVector right;
    right.resize(len - mid_pos);
    int counter = 0;
    for (int i = mid_pos; i < len; i++) {
        right[counter] = d[i];
        
        counter++;
    }

    //cout << "Left half: " << left.AsString() << ", en: " << compute_entropy(left);
    // cout << "\tRight half: " << right.AsString() << ", en: " << compute_entropy(right) << endl;
    
    return (compute_entropy(left) < MIN_WELD_ENTROPY || compute_entropy(right) < MIN_WELD_ENTROPY);
}
        


 

// shadows are contigs that are nearly identical but arise from sequencing error-containing kmers
bool IsShadow(const DNAVector & a, const DNAVector & b, int startA, int startB, int k) 
{
    //return false;
    
    
    int i;
    int n = 0;
    int nn = 0;
    int last = -1;
    int len = 0;
    for (i=startA; i<a.isize(); i++) {
        int x = i-startA + startB;
        
        if (x >= b.isize())
            break;
        len++;
        if (a[i] != b[x]) {     
            //cout << "Mismatch @ " << i << " and " << x << endl;
            if (last >= 0) {
                int dist = i-last;
                if (x > 3 && i > 3 && a[i-1] != b[x-1] && a[i-2] != b[x-2])
                    break;
                if (dist == k+1) {
                    n++;
                } else {
                    nn++;
                }
            }
            last = i;
        } else {
            //cout << i << " and " << x << endl;
        }
    }
    
    int expect = (int)(0.9*(double(len/(k+1)-1)));
    //cout << "Len: " << len << " Expect: " << expect << " Observed: " << n << " diss: " << nn << endl;
    if (n >= expect && n > 4 && nn < n/5) {
        return true;
    }
    return false;
    
}


// e.g., given the string:
// >a1;74093 K: 25 length: 8833
// extract "74093"
double Coverage(const string &s) 
{
    const char *nptr = std::strchr(s.c_str(), ';');
    if (nptr == NULL) // ';' not found
        return 1.0;
    
    char *endptr;
    double ret = std::strtod(++nptr, &endptr);
    if (endptr == nptr || ret < 1.0) // conversion not possible, or value < 1.0
        return 1.0;
    
    return ret;
}

bool IsGoodCoverage(double a, double b, double min_iso_ratio) 
{
    
    if (a > b) {
        double tmp = b;
        b = a;
        a = tmp;
    }
    
    if (a/b > min_iso_ratio) {
        return(true);
    }
    else {
        return(false);
    }
        
    /*  original way
    
    //return true;
    
    double dev_a = sqrt(a);
    double dev_b = sqrt(b);
    
    double mean = (a+b)/2;
    
    if (dev_b > dev_a) {
        double dev_tmp = dev_a;
        dev_a = dev_b;
        dev_b = dev_tmp;
        
        double tmp = a;
        a = b;
        b = tmp;
    }
    
    //cout << "Coverage check: " << a << " mean: " << mean << " dev: " << dev_a << endl;
    
    if ((a - mean) < 10*dev_a)
        return true;
    
    double ratio = a/b;
    if (ratio < 1.)
        ratio = 1./ratio;
    
    if (ratio < 100.)
        return true;
    else
        return false;


    */

}



class Welder
{
public:
    Welder(int k, int kk) {
        m_k = k;
        m_kk = kk;
        m_pTab = NULL;
    }
    
    void SetTable(NonRedKmerTable * p) {
        m_pTab = p;
    }
    

    // constructs the required weldable kmer from two contigs a,b and positions one and two
    void WeldableKmer(DNAVector & out, const DNAVector & a, int one, const DNAVector & b, int two) 
    {
        out.resize(m_kk);
        int flank = (m_kk - m_k)/2;
        
        int startA = one-flank;
        int stopA = one+m_k;
        int startB = two+m_k;
        int stopB = startB + flank;
        
        if (DEBUG) 
            cerr << "weldableKmer(" << startA << "," << stopA << "); (" << startB << "," << stopB << ")" << endl;
        

        if (startA < 0 || stopB >= b.isize()) {
            out.resize(0);
            if (DEBUG) 
                cerr << "range out of bounds" << endl;
            return;
        }
        
        int i;
        int j = 0;
        for (i=startA; i<stopA; i++) {
            out[j] = a[i];
            j++;
        }
        for (i=startB; i<stopB; i++) {
            out[j] = b[i];
            j++;
        }
        
        if (DEBUG)
            cerr << "\tweld candidate: " << out.AsString() << endl;

    }
    
    
    bool Weldable(const DNAVector & a, int one, const DNAVector & b, int two, int thresh, string& welding_kmer, int& weldable_kmer_read_count) 
    {
        int i;
        DNAVector d; // stores the required weldabler kmer of length kk
        WeldableKmer(d, a, one, b, two); // constructs teh weldable kmer, stores in (d)
        if (d.isize() == 0)
            return false;
        
        int count = m_pTab->GetCount(d, 0); // see if the weldable kmer exists among the reads and get read count.
        weldable_kmer_read_count = count;
        

        if (DEBUG) 
            cerr << "got weldable candidate: " << d.AsString() << " with count: " << weldable_kmer_read_count << endl;
        

        if (count >= thresh) {
            
            // cout << d.AsString() << " with count " << count << endl;
            
            welding_kmer = d.AsString();

            return true;
            
        }
        else
            return false;
    }
    
private:
    NonRedKmerTable * m_pTab;
    
    int m_k;
    int m_kk;
};

void Add(vecDNAVector & all, DNAVector & add, int & counter) 
{
    #pragma omp critical
    {        
#ifdef NO_ADD_FIX
#warning add fix NOT activated 
        if (counter >= all.isize()) {
            all.resize(all.isize() + 1000000);
            // cerr << "Resizing to : " << all.isize() << endl;
        }
        all[counter] = add;
        counter++;
#else
#warning add fix ACTIVATED
        all.push_back(add);
        counter++;
#endif
    }
    
}


int main(int argc,char** argv)
{
    
    
    commandArg<string> aStringCmmd("-i","input fasta");
    commandArg<string> readStringCmmd("-r","read fasta");
    commandArg<bool> strandCmmd("-strand","strand specific", false);
    commandArg<int> kCmmd("-k","k-mer size for pooling", 24);
    commandArg<int> kkCmmd("-kk","k-mer size for welding", 48);
    commandArg<int> threadsCmmd("-t", "number of threads (default: use OMP_NUM_THREADS env var)", -1);
    //commandArg<string> bStringCmmd("-o","output fasta");
    commandArg<double> glueFactorCmmd("-glue_factor", "fraction of max (iworm pair coverage) for read glue support", 0.05);
    commandArg<int> minGlueCmmd("-min_glue", "absolute min glue support required", 2);
    commandArg<double> minIsoRatioCmmd("-min_iso_ratio", "min ratio of (iworm pair coverage) for join", 0.05);
    commandArg<double> minKmerEntropyCmmd("-min_kmer_entropy", "min entropy value for kmers", MIN_KMER_ENTROPY); 
    commandArg<double> minWeldEntropyCmmd("-min_weld_entropy", "min entropy value for each half of a welding-(kk)mers", MIN_WELD_ENTROPY); 
    
    commandArg<bool> reportWeldsCmmd("-report_welds", "report the welding kmers", false);
    commandArg<bool> debugCmmd("-debug", "verbosely describes operations", false);

    commandLineParser P(argc,argv);
    P.SetDescription("Makes a graph out of a fasta");
    P.registerArg(aStringCmmd);
    P.registerArg(readStringCmmd);
    P.registerArg(strandCmmd);
    P.registerArg(kCmmd);
    P.registerArg(kkCmmd);
    P.registerArg(threadsCmmd);
    P.registerArg(glueFactorCmmd);
    P.registerArg(minGlueCmmd);
    P.registerArg(minIsoRatioCmmd);
    
    P.registerArg(minKmerEntropyCmmd);
    P.registerArg(minWeldEntropyCmmd);
    
    P.registerArg(reportWeldsCmmd);
    
    P.registerArg(debugCmmd);
    
    //P.registerArg(bStringCmmd);
    
    P.parse();
    
    cerr << "-----------------------------------------" << endl
         << "--- Chrysalis: GraphFromFasta -----------" << endl
         << "-- (cluster related inchworm contigs) ---" << endl
         << "-----------------------------------------" << endl << endl;
    


    string aString = P.GetStringValueFor(aStringCmmd); //inchworm contigs file
    bool sStrand = P.GetBoolValueFor(strandCmmd); // indicates strand-specific mode
    string readString = P.GetStringValueFor(readStringCmmd); // rna-seq reads file (strand-oriented if in strand-specific mode
    int k = P.GetIntValueFor(kCmmd); // kmer size for pooling, must be 24
    int kk = P.GetIntValueFor(kkCmmd); // kmer size for welding, default 48 = kmer + 1/2 kmer on each side of the kmer
    int num_threads = P.GetIntValueFor(threadsCmmd);
    bool report_welds = P.GetBoolValueFor(reportWeldsCmmd);
    double glue_factor = P.GetDoubleValueFor(glueFactorCmmd);
    int min_glue_required = P.GetIntValueFor(minGlueCmmd);
    double min_iso_ratio = P.GetDoubleValueFor(minIsoRatioCmmd);

    MIN_KMER_ENTROPY = P.GetDoubleValueFor(minKmerEntropyCmmd);
    MIN_WELD_ENTROPY = P.GetDoubleValueFor(minWeldEntropyCmmd);
    
    DEBUG = P.GetBoolValueFor(debugCmmd);

    if (min_glue_required < 1) {
        cerr << "-error, cannot have less than 1 read as glue.  Setting -min_glue to 1" << endl;
        min_glue_required = 1;
    }
    

    if (num_threads > 0) {
        cerr << "-setting num threads to: " << num_threads << endl;
        omp_set_num_threads(num_threads);
    }
#ifndef NO_SCHEDULE_FIX
#warning schedule fix ACTIVATED
    int real_num_threads;
    #pragma omp parallel
    {
      #pragma omp single
      {
        real_num_threads = omp_get_num_threads();
      }
    }
    cerr << "-running on " << real_num_threads << " threads" << endl;
#else
#warning schedule fix NOT activated 
#endif

    
    // read inchworm contigs into memory
    vecDNAVector dna;
    
    cerr << "GraphFromFasta: Reading file: " << aString << endl;
#ifdef NO_BIGCHUNK_FIX
#warning bigchunk fix NOT activated 
    dna.Read(aString);
#else    
#warning bigchunk fix ACTIVATED 
    dna.Read(aString, false, false, 1000);
#endif
    cerr << "done!" << endl;
    
    // decode inchworm contigs into kmer composition
    KmerAlignCore core;
    core.SetNumTables(2);
    TranslateBasesToNumberExact trans;
    core.SetTranslator(&trans);
    core.AddData(dna);
    core.SortAll();
    
    int i, j;
    
    if (k != 24) {
        cerr << "The only size of k supported is 24! Exiting!" << endl;
        return -1;
    }
        
    
    bool bNoWeld = false;
    
    NonRedKmerTable kmers(kk);
    
    
    //string dbname;
    //TempFile(dbname, aString);
    //kmers.Read(dbname);
    //kmers.Add(seq);
    
    
    Welder weld(k, kk); // decides if read support exists to weld two inchworm contigs together into the same component.
    
    
    int max_count = 10;
    
    int count = 0;


    // Get all potential welding kmers by all-vs-all search among the inchworm contigs    
    vecDNAVector crossover;
    //crossover.reserve(100000);
    cerr << "Collecting (kk)-mers needed for welding..." << endl;
    int counter = 0;
    
    double start_time = omp_get_wtime();

    int iworm_counter = 0;
    int schedule_chunksize = 1000;
#ifndef NO_SCHEDULE_FIX
    schedule_chunksize = dna.isize() / real_num_threads / 100;
    if(schedule_chunksize<1) schedule_chunksize=1;
    cerr << "-setting omp for schedule chunksize to " << schedule_chunksize << " for " << dna.isize() << " iworm contigs" << endl;
#endif

    #pragma omp parallel for schedule(dynamic,schedule_chunksize) private(j)
    for (i=0; i<dna.isize(); i++) {
        DNAVector & d = dna[i]; // inchworm contig [i]
        
        #pragma omp atomic
        iworm_counter++;

        if (iworm_counter % 1000 == 0 || iworm_counter == dna.isize()-1) {
            #pragma omp critical
            cerr << "\rProcessed: " << iworm_counter/(double)dna.isize()*100 << " % of iworm contigs.    ";
        }
        
        for (j=0; j<=d.isize()-k; j++) {
            DNAVector sub; // a kmer
            sub.SetToSubOf(d, j, k);
            
            if (IsSimple(sub))
                continue; // ignore kmers that are low complexity
            

            // find other inchworm contigs that have kmer matches
            svec<KmerAlignCoreRecord> matchesFW, matchesRC;   
            
            core.GetMatches(matchesFW, sub);
            if (!sStrand) {
                sub.ReverseComplement();
                core.GetMatches(matchesRC, sub);
            }
            
            int x;
            
            for (x=0; x<matchesFW.isize(); x++) {
                int c = matchesFW[x].GetContig();
                if (c == i)
                    continue;  // matches itself
                
                DNAVector & dd = dna[c];	
                int start = matchesFW[x].GetPosition();
                
                if (DEBUG) {
                    cerr << "(Phase1: kmer match between " << dna.Name(i) << " and " << dna.Name(c)
                         << " at F positions: " << j << " and " << start << endl;
                } 

                DNAVector add;
                
                weld.WeldableKmer(add, d, j, dd, start);
                if (add.isize() > 0 && ! IsSimple(add) && (! SimpleHalves(add) )) {
                    Add(crossover, add, counter);      
                }
                
                weld.WeldableKmer(add, dd, start, d, j);
                if (add.isize() > 0 && (! IsSimple(add)) && (! SimpleHalves(add) )) {
                    Add(crossover, add, counter);
                }
            }
            for (x=0; x<matchesRC.isize(); x++) {
                int c = matchesRC[x].GetContig();
                if (c == i)
                    continue;
                DNAVector dd = dna[c];
                dd.ReverseComplement();
                
                int start = dd.isize() - matchesRC[x].GetPosition() - k;
                DNAVector add;
                
                weld.WeldableKmer(add, d, j, dd, start); 
                if (add.isize() > 0 && (! IsSimple(add)) && (! SimpleHalves(add)))
                    Add(crossover, add, counter);	 
                
                
                weld.WeldableKmer(add, dd, start, d, j);
                if (add.isize() > 0 && ! IsSimple(add) && (! SimpleHalves(add)))
                    Add(crossover, add, counter);	 
            }
        }
    }
    
    double end_time = omp_get_wtime();
    cerr << endl << endl << "...done (" << end_time - start_time << " seconds)" << endl;
    

#ifdef NO_BIGCHUNK_FIX
    crossover.resize(counter);
#endif
    
    //cerr << "Setting up/sorting structure." << endl;
    kmers.SetUp(crossover);
    
    if (!bNoWeld) {
        cerr << "Setting up reads for streaming..." << endl;
#ifdef NO_FASTSTREAM
#warning DNAStringStreamFast NOT activated
        vecDNAVectorStream seq;
#else
#warning DNAStringStreamFast ACTIVATED
        DNAStringStreamFast seq;
#endif
        seq.ReadStream(readString);  //readString is the name of the file containing the reads
        
        cerr << "Identifying reads that support welding of iworm contigs..." << endl;
        kmers.AddData(seq);
        cerr << endl << "Done!" << endl;
        
        weld.SetTable(&kmers);
    }
    
    //=================================================================
    // Cluster the inchworm contigs based on read-weld support

    svec<int> mapped;
    mapped.resize(dna.isize(), -1); // init all mapped enrtries to -1 (unmapped).  Stores Pool IDs for those clustered.
    svec<Pool> pool; // a vector of vector<int>
    
    start_time = omp_get_wtime();
    
    // init pools and locks
    // Create an array of locks; one for each element of mapped[]
    // omp_lock_t *mapped_lock = new omp_lock_t[dna.isize()];
    for (i=0; i<dna.isize(); i++) {
        // omp_init_lock(&mapped_lock[i]);
        
        Pool tmp;
        // add a pool to the list of pools, and init with current iworm contig
        pool.push_back(tmp);
        int poolID = pool.isize()-1;
        pool[poolID].add(i);
                                                                        \
        mapped[i] = poolID;  // store pool ID that corresponds to iworm contig (i)
        
        if (i != poolID) {
            cerr << "Error, initial poolID != iworm id"; // shouldn't happen
            exit(2);
        }
        
    }

    vector<bool> toasted;
    toasted.resize(dna.isize(), false);
    
    #pragma omp parallel for schedule(dynamic, schedule_chunksize) private(j)
    for (i=0; i<dna.isize(); i++) {

        if (DEBUG) 
            cerr << "Phase1: iworm(i) = " << dna.Name(i) << endl;
        

        int cutoff = 0;
        DNAVector & d = dna[i];
        if (d.isize() < k) {
            
            if (DEBUG) {
                cerr << "ignoring: " << dna.Name(i) << " since less than length: " << k << endl;
            }
                
            continue;
        }
        
        if (mapped[i] == -2) { //shadow
            if (DEBUG) {
                cerr << "ignoring: " << dna.Name(i) << " since flagged as a shadow contig " << endl;
            }
            
            continue;
        }
                

        // iterate through kmers of inchworm contig
        for (j=0; j<=d.isize()-k; j++) {
            
            DNAVector sub;
            sub.SetToSubOf(d, j, k);
            
            svec<KmerAlignCoreRecord> matchesFW, matchesRC;   
            
            core.GetMatches(matchesFW, sub);
            if (!sStrand) {
                sub.ReverseComplement();
                core.GetMatches(matchesRC, sub);
            }
            
            if (IsSimple(sub) && matchesFW.isize() + matchesRC.isize() > 1) {
                if (DEBUG) {
                    cerr << "kmer: " << sub.AsString() << " ignored since either low complex and too many iworm matches"<< endl;
                }

                continue;      
            }
            
            int x;
            
            double coverage = Coverage(dna.Name(i));
                        
            if (DEBUG) {
                cerr << "kmer: " << sub.AsString() << " has " << matchesFW.isize() << " F matches and "
                     << matchesRC.isize() << " R iworm contig matches." << endl;
            }


            // Process forward matching kmers
            for (x=0; x<matchesFW.isize(); x++) {
                int c = matchesFW[x].GetContig();
                
                if (DEBUG) {
                    cerr << "kmer: " << sub.AsString() << " supports match between " << dna.Name(i) << " and " << dna.Name(c) << endl;
                }
                if (c == i) {
                    if (DEBUG) {
                        cerr << "-ignoring self iworm match." << endl;
                    }
                    
                    continue;
                }
                
                double coverage_other = Coverage(dna.Name(c));
                
    
                double higher_coverage_val = (coverage > coverage_other) ? coverage : coverage_other;
                int minCov = (int) (higher_coverage_val * glue_factor);
                if (minCov < min_glue_required) {
                    minCov = min_glue_required;
                }
                
                DNAVector & dd = dna[c];	
                int start = matchesFW[x].GetPosition();
                
                
                if (DEBUG) {
                    cerr << "Phase2: kmer match between " << dna.Name(i) << " and " << dna.Name(c)
                         << " at F positions: " << j << " and " << start << endl;
                } 
                
                
                //cout << "TEST_WELD: " << I << " " << J << " ";
                string welding_kmer;
                int welding_kmer_read_count;
                if (!bNoWeld 
                    && 
                    !(weld.Weldable(d, j, dd, start, minCov, welding_kmer, welding_kmer_read_count) 
                      || 
                      weld.Weldable(dd, start, d, j, minCov, welding_kmer, welding_kmer_read_count))) {
                    
                    if (DEBUG) {
                        cerr << "\t no welding kmer avail " << welding_kmer << ", " << welding_kmer_read_count << endl;
                    }
                    
                    continue;
                }
                                
                if (IsShadow(d, dd, j, start, k) && coverage > 2*coverage_other) {
                    cerr << "Toasting shadow: " << dna.Name(c) << endl;
                    // mapped[c] = -2;
                    #pragma omp critical
                    toasted[c] = true;
                    
                    continue;
               
                } else if (!IsGoodCoverage(coverage, coverage_other, min_iso_ratio)) {
                    //cerr << "Rejecting fw merge between " << dna.Name(i);
                    //cerr << " and " << dna.Name(c) << endl;
                    
                    continue;
                }
                // cerr << "Accept (fw)!!" << endl;
                
                #pragma omp critical
                pool[i].add(c);
                
                if (report_welds) {
                    #pragma omp critical
                    cout << "#Welding: " << dna.Name(i) << " to " << dna.Name(c) 
                         << " with " << welding_kmer << " found in " << welding_kmer_read_count << " reads" << endl;
                }


            }

            if (sStrand) {
                // only doing the forward matches

                if (DEBUG) {
                    cerr << " only procesing forward strand " << endl;
                }
                

                continue; // superfluous since we didn't capture any rc matches
            }
            
            // Process the RC matches now
            for (x=0; x<matchesRC.isize(); x++) {
                int c = matchesRC[x].GetContig();


                if (DEBUG) {
                    cerr << "RCkmer: " << sub.AsString() << " supports match between " << dna.Name(i) << " and " << dna.Name(c) << endl;
                }

                if (c == i) {
                    if (DEBUG) {
                        cerr << "-ignoring self iworm match." << endl;
                    }
                    
                    continue;
                }
                                
                double coverage_other = Coverage(dna.Name(c));
                
                double higher_coverage_val = (coverage > coverage_other) ? coverage : coverage_other;
                int minCov = (int) (higher_coverage_val * glue_factor);
                if (minCov < min_glue_required) {
                    minCov = min_glue_required;
                }
                
                DNAVector dd = dna[c];
                dd.ReverseComplement();  // revcomp a copy of the iworm[c] sequence
                
                int start = dd.isize() - matchesRC[x].GetPosition() - k;  // reverse-complement the match coordinate


                if (DEBUG) {
                    cerr << "kmer match between " << dna.Name(i) << " and " << dna.Name(c)
                         << " at R positions: " << j << " and " << start << endl;
                } 
                
                
                string welding_kmer;
                int welding_kmer_read_count;
                if (!bNoWeld 
                    && 
                    !(weld.Weldable(d, j, dd, start, minCov, welding_kmer, welding_kmer_read_count) 
                      || 
                      weld.Weldable(dd, start, d, j, minCov, welding_kmer, welding_kmer_read_count))) {
                    continue;
                }
                                
                if (IsShadow(d, dd, j, start, k) && coverage > 2*coverage_other) {
                    cerr << "Toasting shadow: " << dna.Name(c) << endl;

                    #pragma omp critical
                    toasted[c] = true;
                    
                    continue;

                } else if (!IsGoodCoverage(coverage, coverage_other, min_iso_ratio)) {
                    // cerr << "Rejecting rc merge between " << dna.Name(i);
                    // cerr << " and " << dna.Name(c) << endl;
                    continue;
                }
                //cerr << "Accept (rc)!!" << endl;


                
                #pragma omp critical (pool)
                {
                    //dna[c].ReverseComplement();     // note: TranscriptomeGraph() of Chrysalis requires inchworm contigs to be properly oriented with respect to each other; now, instead, using new FastaToDeBruijnGraph code that has stranded modes of operation.
                    
                    pool[i].add(c);
                }

                if (report_welds) {
                    #pragma omp critical
                    cout << "#Welding: " << dna.Name(i) << " to " << dna.Name(c) 
                         << " with " << welding_kmer << " found in " << welding_kmer_read_count << " reads" << endl;
                }
                
                //cout << "Mapped sequence " << dna.NameClean(c) << " to pool " << mapped[c] << " -" << endl;
            }
            
        }
    }

    
    end_time = omp_get_wtime();


    // ----------------------------------------------------------------------------- //
    // ------   Transitive Closure, Define Inchworm Clustered Contigs -------------- //
    // ----------------------------------------------------------------------------- //
    
    
    int round = 0;
    bool remapped_flag = true;
    while (remapped_flag) {
        round++;
        cerr << "Transitive closure, round: " << round << endl;
        remapped_flag = false;
        // do transitive closure on the poolings
        svec<int> remapped = mapped;
        
        // trickle pool mappings upward
        
        for (i=0; i<pool.size(); i++) {
            int oldpool_id = i;
            
            Pool& oldpool = pool[oldpool_id];
            
            int lowest_pool_map_id = oldpool_id;
            bool pool_reassignment_required = false;
            for (j=0; j<oldpool.size(); j++) {
                int iworm_id = oldpool.get(j);
                int pool_mapping = remapped[iworm_id];
                
                
                if (pool_mapping != lowest_pool_map_id) {
                    pool_reassignment_required = true;
                }
                if (pool_mapping >= 0 && pool_mapping < lowest_pool_map_id) {
                    lowest_pool_map_id = pool_mapping;
                }
            }
            
            // reassign all members to lowest_pool_map_id (if same as pool_id, some entries may still be higher even if not less)
            if (pool_reassignment_required) {
                for (j=0; j<oldpool.size(); j++) {
                    int iworm_id = oldpool.get(j);
                    remapped[iworm_id] = lowest_pool_map_id;
                    // cerr << "-remapping: iworm(" << iworm_id << ") to pool " << lowest_pool_map_id << endl; 
                }
            }
            
            if (lowest_pool_map_id < oldpool_id) {
                // reassign to lower pool
                Pool tmp;
                for (j=0; j<oldpool.size(); j++) {
                    int iworm_id = oldpool.get(j);
                    pool[lowest_pool_map_id].add(iworm_id);
                    //#pragma omp critical
                    //cout << "RELOCATING: " << iworm_id << " from old pool " << oldpool_id << " to new pool " << lowest_pool_map_id << endl;
                }
                //cerr << "-clearing out old pool: " << oldpool_id << endl;
                pool[oldpool_id] = tmp;
                remapped_flag = true;
            }    
        }
        mapped = remapped; // new assignments
    }
    
    
    
    cerr << "...done (" << end_time - start_time << " seconds)" << endl;

    //------------------------------------------
    // Clustered inchworm contigs now defined.
    // Generate final output
    
    for (i=0; i<pool.isize(); i++) {
        Pool & p = pool[i];
        
        if (p.size() == 0) { continue; }
        
        p.sortvec();

        vector<int> nr_entries; // remove the redundant entries in p
        for (j=0; j<p.size(); j++) {
            int z = p.get(j);
            
            if (toasted[z]) {
                continue;
            }
            

            if (j > 0 && p.get(j-1) == z) {
                // same entry ended up on the pool vec, already reported it.
                continue;
            }
            
            nr_entries.push_back(z);
        }

        stringstream pool_info;                
        pool_info << "#POOL_INFO\t" << i << ":" << "\t";

        cout << "COMPONENT " << i << "\t" << nr_entries.size() << endl;
        for (j = 0; j < nr_entries.size(); j++) {
            int z = nr_entries[j];
                                
            pool_info << z << " ";
            cout << ">Component_" << i << " " << p.size() << " " << z << " [iworm" << dna.Name(z) << "]" << endl;
            PrintSeq(dna[z]);
        }
        pool_info << endl;
        cout << pool_info.str();
        
        cout << "END" << endl;
    }
    
    return 0;
    
}
  
